#region MyJScript - LGPL Copyright (c) 2008 Lionel Lask
//
// This file is part of MyJScript.
//
// MyJScript is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// MyJScript is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with MyJScript; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
//================================================================================
#endregion

using System;
using System.Collections.Generic;
using System.Reflection;
using System.Text;
using Microsoft.Scripting;
using Microsoft.Scripting.Ast;
using Microsoft.Scripting.Actions;
using Microsoft.Scripting.Runtime;

using MyJScript.Runtime;

namespace MyJScript.Compiler
{
    /// <summary>
    /// Code generator for MyJScript
    /// </summary>
    public class Generator
    {
        /// <summary>
        /// Current generated code block
        /// </summary>
        public LambdaBuilder Result
        {
            get { return generatedBlock; }
        }
        private LambdaBuilder generatedBlock;

        /// <summary>
        /// Block for current method and previous one
        /// </summary>
        private LambdaBuilder currentMethod;
        private LambdaBuilder previousMethod;

        /// <summary>
        /// Local variables
        /// </summary>
        private Dictionary<String, Variable> localVariables;

        /// <summary>
        /// Global variables (including functions)
        /// </summary>
        private Dictionary<String, Variable> globalVariables;

        /// <summary>
        /// Statements found
        /// </summary>
        private List<Expression> statements;

        /// <summary>
        /// Block limit in statements list
        /// </summary>
        static private SourceLocation dummyLocation = new SourceLocation(255, 255, 255);
        static private SourceSpan dummySpan = new SourceSpan(dummyLocation, dummyLocation);
        private Expression blockLimit = Ast.Empty(dummySpan);

        /// <summary>
        /// Function call values
        /// </summary>
        private List<Expression> exprs;

        /// <summary>
        /// Expr limit in expression list
        /// </summary>
        private Expression exprLimit = Ast.DebugMark(Ast.Constant(255), "<dummy>");

        /// <summary>
        /// Object to report error
        /// </summary>
        private Report report;

        /// <summary>
        /// Current location
        /// </summary>
        private Location location;

        /// <summary>
        /// Current source span
        /// </summary>
        private SourceSpan span
        {
            get { return new SourceSpan(location.SourceLocation, location.SourceLocation); }
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="reportObject">object to report error</param>
        /// <param name="lexerLocation">location of current token</param>
        public Generator(Report reportObject, Location lexerLocation)
        {
            report = reportObject;
            location = lexerLocation;
            globalVariables = new Dictionary<string, Variable>();
            localVariables = new Dictionary<string, Variable>();
            statements = new List<Expression>();
            exprs = new List<Expression>();
            currentMethod = null;
            previousMethod = null;
        }

        /// <summary>
        /// Initialize generator
        /// </summary>
        private void Initialize()
        {
            generatedBlock = Ast.Lambda("MyJScript", typeof(object));
        }

        /// <summary>
        /// Start a MyJScript program
        /// </summary>
        public void BeginProgram()
        {
            // Create Namespace and Class
            Initialize();
        }

        /// <summary>
        /// End a MyJScript program
        /// </summary>
        public void EndProgram()
        {
            // Create statements
            generatedBlock.Body = Ast.Block(statements);

            // Declare global variables
            globalVariables.Clear();
        }

        /// <summary>
        /// Start a function declaration
        /// </summary>
        /// <param name="name">function name</param>
        /// <param name="arguments">parameters name</param>
        public void BeginFunction(String name, List<String> arguments)
        {
            // This function already exist
            if (name != null && globalVariables.ContainsKey(name))
            {
                report.Error(location, Report.ProcedureAlreadyDefinedError, String.Format("{0} is already defined", name));
                return;
            }

            // Create a block for the function
            previousMethod = currentMethod;
            currentMethod = Ast.Lambda(name ?? "<member function>", typeof(object));

            // Create dictionnary for local variables
            localVariables = new Dictionary<string, Variable>();

            // Declare this as parameter
            currentMethod.CreateParameter(SymbolTable.StringToId("this"), typeof(object));
            
            // Create a new method
            if (arguments.Count > 0)
            {
                // Check that each parameter name is unique
                bool uniqueParameter = true;
                int length = arguments.Count;
                for (int i = 0; i < length; i++)
                    for (int j = 0; j < length; j++)
                        if (i != j && arguments[i].Equals(arguments[j]))
                            uniqueParameter = false;
                if (!uniqueParameter)
                    report.Error(location, Report.ParameterNameMustBeUniqueError, String.Format("Parameter name must be unique for function \"{0}\"",name));

                // Declare parameters
                foreach (string parameter in arguments)
                    currentMethod.CreateParameter(SymbolTable.StringToId(parameter), typeof(object));
            }
        }

        /// <summary>
        /// End a function declaration
        /// </summary>
        /// <returns>block for function</returns>
        public CodeBlockExpression EndFunction()
        {
            // HACK: Add a "return this;" statement when function is used as constructor
            statements.Add(Ast.Return(Ast.Read(currentMethod.Parameters[0])));

            // Add all statements
            Expression blockStatements = EndBlock();
            currentMethod.Body = blockStatements;
            LambdaExpression result = currentMethod.MakeLambda();

            // Not in a method now
            currentMethod = previousMethod;
            previousMethod = null; // HACK: work only if no more than two are imbricated

            // Return code as expression
            return Ast.CodeBlockExpression(result, typeof(MyJScriptCallTarget));
        }

        /// <summary>
        /// Start a new block in program
        /// </summary>
        public void BeginBlock()
        {
            statements.Add(blockLimit);
        }

        /// <summary>
        /// End a block
        /// </summary>
        /// <returns>all statements for the block</returns>
        public Expression EndBlock()
        {
            // Look for the begin of block
            int count = statements.Count;
            int first;
            for (first = count - 1; first >= 0; first--)
                if (statements[first] == blockLimit)
                    break;

            // Create an array with all statements
            List<Expression> blockStatements = new List<Expression>();
            for (int i = first + 1; i < count; i++)
                blockStatements.Add(statements[i]);

            // Remove all statements of the block including block limit
            for (int i = first; i < count; i++)
                statements.RemoveAt(statements.Count - 1);

            // Return block
            return Ast.Block(blockStatements);
        }

        /// <summary>
        /// Add statement to current block or to main depending on context.
        /// </summary>
        /// <param name="statement">statement to add</param>
        public void AddStatement(Expression statement)
        {
            // Add statement to list
            statements.Add(statement);
        }

        /// <summary>
        /// Start a list of expression.
        /// </summary>
        public void BeginListExpr()
        {
            exprs.Add(exprLimit);
        }

        /// <summary>
        /// End a list of expression.
        /// </summary>
        /// <returns>the list for this level</returns>
        public List<Expression> EndListExpr()
        {
            // Look for the begin of list
            int count = exprs.Count;
            int first;
            for (first = count - 1; first >= 0; first--)
                if (exprs[first] == exprLimit)
                    break;

            // Create a list with all expressions
            List<Expression> currentList = new List<Expression>();
            for (int i = first + 1; i < count; i++)
                currentList.Add(exprs[i]);

            // Remove all statements of the block including block limit
            for (int i = first; i < count; i++)
                exprs.RemoveAt(exprs.Count - 1);

            return currentList;
        }

        /// <summary>
        /// Add expression in current list.
        /// </summary>
        /// <param name="expr">expression to add</param>
        public void AddExpression(Expression expr)
        {
            exprs.Add(expr);
        }

        /// <summary>
        /// Import a .NET library.
        /// </summary>
        /// <param name="name">library name</param>
        /// <returns>statement to import</returns>
        public Expression UsingLibrary(String name)
        {
            // Call expression
            Expression callExpr = Ast.Call(
                typeof(MJSContext).GetMethod("UseLibrary"),
                Ast.CodeContext(),
                Ast.Constant(name)
            );

            // Set global var
            return DeclareGlobalVar(name, callExpr);          
        }

        /// <summary>
        /// Declare a new variable (local or global depending on current context).
        /// </summary>
        /// <param name="name">variable name</param>
        /// <param name="initValue">init value</param>
        /// <returns>the assign statement</returns>
        public Expression DeclareVariable(string name, Expression initValue)
        {
            // Declare locally if possible
            if (currentMethod != null)
                return DeclareLocalVar(name, initValue);
            else
                return DeclareGlobalVar(name, initValue);
        }

        /// <summary>
        /// Declare a new global variable.
        /// </summary>
        /// <param name="name">variable name</param>
        /// <param name="initValue">init value</param>
        /// <returns>Statement to set the initial value</returns>
        public Expression DeclareGlobalVar(string name, Expression initValue)
        {
            // Ignore if already exist, generate set value
            if (globalVariables.ContainsKey(name))
                return AssignGlobalVariable(globalVariables[name], initValue);

            // Create a new global variable
            Variable variable = generatedBlock.CreateLocalVariable(
                    SymbolTable.StringToId(name),
                    typeof(object));

            // Add to current declaration
            globalVariables.Add(name, variable);
            
            // Generate set initial value statement
            return AssignGlobalVariable(variable, initValue);
        }

        /// <summary>
        /// Declare a new function.
        /// </summary>
        /// <param name="name">function name</param>
        /// <param name="impl">implementation for function</param>
        /// <returns>statement</returns>
        public Expression DeclareFunction(CodeBlockExpression impl)
        {
            // Look if already exist
            Variable variable;
            string name = impl.Block.Name;
            if (globalVariables.ContainsKey(name))
                variable = globalVariables[name];

            // Else create a new variable 
            else
            {
                // Create a new global variable
                variable = generatedBlock.CreateLocalVariable(
                        SymbolTable.StringToId(name),
                        typeof(object));

                // Add to current declaration
                globalVariables.Add(name, variable);
            }

            // Generate set a global variable with the function
            List<Expression> statements = new List<Expression>();
            statements.Add(
                Ast.Statement(span,
                    Ast.Assign(
                        variable,
                        impl
                    )
                )
            );

            // Generate set member on global context
            statements.Add(
                AssignMember(
                    Ast.Call(
                        typeof(MJSContext).GetMethod("GetContext", new Type[0])
                    ),
                    name,
                    Ast.Read(variable)
                )
            );

            return Ast.Block(statements);
        }

        /// <summary>
        /// Declare a new local variable.
        /// </summary>
        /// <param name="name">variable name</param>
        /// <param name="initValue">init value</param>
        /// <returns>Statement to set the initial value</returns>
        public Expression DeclareLocalVar(string name, Expression initValue)
        {
            // Ignore if already exist, generate set value
            if (localVariables.ContainsKey(name))
                return AssignVariable(localVariables[name], initValue);

            // Ignore if a parameter has same name
            foreach (Variable param in currentMethod.Parameters)
                if (SymbolTable.IdToString(param.Name).Equals(name))
                {
                    // Generate set value if value is not null
                    if (initValue != null)
                        return AssignVariable(param, initValue);
                    else
                        return Ast.Empty();
                }

            // Create a new variable
            Variable variable = currentMethod.CreateLocalVariable(
                    SymbolTable.StringToId(name),
                    typeof(object));

            // Add to current declaration
            localVariables.Add(name, variable);

            // Generate set initial value statement
            return AssignVariable(variable, initValue);
        }

        /// <summary>
        /// Get a variable reference.
        /// </summary>
        /// <param name="name">variable name</param>
        /// <returns>reference expression</returns>
        public Expression Variable(string name)
        {
            // Look first in local variables
            Variable variable = null;
            if (localVariables.ContainsKey(name))
                variable = localVariables[name];

            // Else look in parameter and global variable
            else 
            {
                // Parameter store only a value
                if (currentMethod != null)
                {
                    // Look if it's a parameter
                    foreach (Variable param in currentMethod.Parameters)
                        if (SymbolTable.IdToString(param.Name).Equals(name))
                            variable = param;

                    // Then, return only the value
                    if (variable != null)
                        return Ast.Read(variable);
                }

                // Try a global variable
                if (variable == null)
                {
                    // Look for a global variable 
                    if (!globalVariables.ContainsKey(name))
                    {
                        // Leave a chance to DLR
                        return Ast.Read(SymbolTable.StringToId(name));
                    }

                    // Global variable found
                    variable = globalVariables[name];
                }
            }

            // Return value of variable
            return Ast.Read(variable);
        }

        /// <summary>
        /// Get value for member of an instance.
        /// </summary>
        /// <param name="instance">instance</param>
        /// <param name="name">member name</param>
        /// <returns>member value</returns>
        public Expression Member(Expression instance, String name)
        {
            return Ast.Action.GetMember(
                SymbolTable.StringToId(name),
                typeof(object),
                instance
            );
        }

        /// <summary>
        /// Create a string constant from a value.
        /// </summary>
        /// <param name="value">value</param>
        /// <returns>expression</returns>
        public Expression StringConstant(String value)
        {
            return Ast.Constant(value, typeof(string));
        }

        /// <summary>
        /// Create an int constant from a value.
        /// </summary>
        /// <param name="value">value</param>
        /// <returns>expression</returns>
        public Expression NumberConstant(int value)
        {
            return Ast.Constant(value, typeof(int));
        }

        /// <summary>
        /// Create a statement to assign a value to a variable.
        /// </summary>
        /// <param name="name">name of variable</param>
        /// <param name="value">value to assign</param>
        /// <returns>the statement</returns>
        public Expression AssignVariable(String name, Expression value)
        {
            // Look in local variable to find Variable object
            if (localVariables.ContainsKey(name))
                return AssignVariable(localVariables[name], value);

            // If in a function
            if (currentMethod != null)
            {
                // Look if it's a parameter
                foreach (Variable param in currentMethod.Parameters)
                    if (SymbolTable.IdToString(param.Name).Equals(name))
                        return AssignVariable(param, value);
            }

            // Else look in global variable to find Variable object
            if (globalVariables.ContainsKey(name))
                return AssignGlobalVariable(globalVariables[name], value);

            // Finally, declare then assign value
            return DeclareVariable(name, value);
        }

        /// <summary>
        /// Assign a value to a global variable.
        /// </summary>
        /// <param name="variable">variable</param>
        /// <param name="value">new value</param>
        /// <returns>statement to assign the value</returns>
        public Expression AssignGlobalVariable(Variable variable, Expression value)
        {
            // Assign to the global variable
            List<Expression> statements = new List<Expression>();
            statements.Add(
                AssignVariable(variable, value)
            );

            // Set in MyJScript context
            statements.Add(
                AssignMember(
                    Ast.Call(
                        typeof(MJSContext).GetMethod("GetContext", new Type[0])
                    ),
                    SymbolTable.IdToString(variable.Name),
                    Ast.Read(variable)
                )
            );

            return Ast.Block(statements);
        }

        /// <summary>
        /// Create a statement to assign a value to a variable.
        /// </summary>
        /// <param name="variable">variable reference</param>
        /// <param name="value">value to assing</param>
        /// <returns>the statement</returns>
        public Expression AssignVariable(Variable variable, Expression value)
        {
            // Get the right constructor: with or without value
            if (value == null)
            {
                // Assign to null
                return Ast.Statement(span,
                    Ast.Assign(
                        variable,
                        Ast.Null(typeof(object))
                    )
                );
            }
            else
            {
                // Assign to variable
                return Ast.Statement(span,
                    Ast.Assign(
                        variable,
                        Ast.Convert(   // Need for ValueType
                            value,
                            typeof(object)
                        )
                    )
                );
            }
        }

        /// <summary>
        /// Create a statement to assign a value to a member.
        /// </summary>
        /// <param name="instance">instance</param>
        /// <param name="name">member name</param>
        /// <param name="value">new value</param>
        /// <returns>assign statement</returns>
        public Expression AssignMember(Expression instance, String name, Expression value)
        {
            return Ast.Statement(span,
                Ast.Action.SetMember(
                    SymbolTable.StringToId(name),
                    typeof(object),
                    instance,
                    value
                )
            );
        }

        /// <summary>
        /// Create a function call expression with its parameters.
        /// </summary>
        /// <param name="function">name of function to call</param>
        /// <param name="values">values for function arguments</param>
        /// <returns>the call expression</returns>
        public Expression FunctionCall(String function, List<Expression> values)
        {
            // Build an array for function delagate, parameters and this
            int length = values.Count;
            Expression[] array = new Expression[length+2];

            // Get delegate
            array[0] = Ast.Read(SymbolTable.StringToId(function));

            // Add this
            array[1] = Ast.Call(
                typeof(MJSContext).GetMethod("GetContext", BindingFlags.Static | BindingFlags.Public)
            );

            // Add other parameters
            for (int i = 0; i < length; i++)
                array[i+2] = values[i];

            // Call method 
            return Ast.Action.Call(
                typeof(object),
                array
            );
        }

        /// <summary>
        /// Create a method call expression with its parameters
        /// using an instance.
        /// </summary>
        /// <param name="function">name of function to call</param>
        /// <param name="values">values for function arguments</param>
        /// <returns>the call expression</returns>
        public Expression MethodCall(Expression instance, String function, List<Expression> values)
        {
            // Build an array for function delagate, parameters and instance
            int length = values.Count;
            Expression[] array = new Expression[length+1];

            // Add instance as parameter for invoke call
            array[0] = instance;

            // Add other parameters
            for (int i = 0; i < length; i++)
                array[i+1] = values[i];

            // Invoke member
            return Ast.Action.InvokeMember(
                SymbolTable.StringToId(function),
                typeof(object),
                InvokeMemberActionFlags.None,
                new CallSignature(values.Count),
                array
                );
        }

        /// <summary>
        /// Create a new empty instance.
        /// </summary>
        /// <returns>a new object</returns>
        public Expression NewObject()
        {
            return Ast.New(
                typeof(MJSObject).GetConstructor(new Type[0])
            );
        }

        /// <summary>
        /// Create a new instance.
        /// </summary>
        /// <param name="className">class name</param>
        /// <param name="values">constructor parameters</param>
        /// <returns>a new object</returns>
        public Expression NewObject(String className, List<Expression> values)
        {
            // Create a new object           
            Expression newObject = NewObject();

            // Build an array for function delagate, parameters and this
            int length = values.Count;
            Expression[] array = new Expression[length + 2];

            // Get delegate 
            array[0] = Ast.Read(SymbolTable.StringToId(className));

            // Add this
            array[1] = newObject;

            // Add other parameters
            for (int i = 0; i < length; i++)
                array[i + 2] = values[i];

            // Call method 
            return Ast.Action.Call(
                typeof(object),
                array
            );
        }

        /// <summary>
        /// Create a statement from an expression.
        /// </summary>
        /// <param name="value">an expression</param>
        /// <returns>a statement</returns>
        public Expression Statement(Expression value)
        {
            return Ast.Statement(span, value);
        }

        /// <summary>
        /// Create a while statement.
        /// </summary>
        /// <param name="condition">end condition</param>
        /// <param name="toRepeat">statements to repeat</param>
        /// <returns>a while statement</returns>
        public Expression While(Expression condition, Expression toRepeat)
        {
            // Create a new iteration statement
            return Ast.While(Ast.Convert(condition, typeof(bool)), toRepeat, Ast.Empty());
        }

        /// <summary>
        /// Create a return statement with or without a value.
        /// </summary>
        /// <param name="value">value to return</param>
        /// <returns>a return statement</returns>
        public Expression Return(Expression value)
        {
            // Can't return at global scope
            if (currentMethod == null)
            {
                report.Error(location, Report.ReturnAtGlobalScopeError, "You can't use \"return\" at global scope");
                return Ast.Empty();
            }
            
            // Create return statement
            if (value != null)
                return Ast.Return(value);
            else
                return Ast.Return(Ast.Null(typeof(object)));
        }

        /// <summary>
        /// Create an if/else statement.
        /// </summary>
        /// <param name="condition">condition</param>
        /// <param name="ifTrue">statements list to execute if condition is true</param>
        /// <param name="ifFalse">statements list to execute if condition is false</param>
        /// <returns>a condition statement</returns>
        public Expression IfElse(Expression condition, Expression ifTrue, Expression ifFalse)
        {
            // No else
            if (ifFalse == null)
                return Ast.IfThen(
                    Ast.Convert(condition, typeof(bool)),
                    ifTrue
                );

            // Create an if-then-else statement
            return Ast.IfThenElse(
                Ast.Convert(condition, typeof(bool)),
                ifTrue,
                ifFalse
            );
        }

        /// <summary>
        /// Create a binary expression from two expression.
        /// </summary>
        /// <param name="first">first expression</param>
        /// <param name="optype">operator</param>
        /// <param name="second">second expression</param>
        /// <returns></returns>
        public Expression BinaryOperator(Expression first, Operators optype, Expression second)
        {
            return Ast.Action.Operator(optype, typeof(object), first, second);
        }
    }
}
