#region MyJScript - LGPL Copyright (c) 2008 Lionel Lask
//
// This file is part of MyJScript.
//
// MyJScript is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// MyJScript is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with MyJScript; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
//================================================================================
#endregion

using System;
using System.Text;
using Microsoft.Scripting;
using Microsoft.Scripting.Hosting;

namespace MyJScript.Compiler
{
	/// <summary>
	///   This class is used to report errors and warnings t te user.
	/// </summary>
	public class Report
	{
		/// <summary>
		/// Lexical error
		/// </summary>
		public const int LexicalError = 100;

        /// <summary>
        /// Unexpected end-of-file error
        /// </summary>
        public const int UnexpectedEOFError = 200;
			
		/// <summary>
		/// Syntax error
		/// </summary>
		public const int SyntaxError = 300;

        /// <summary>
        /// Parameter must be unique error
        /// </summary>
        public const int ParameterNameMustBeUniqueError = 400;

        /// <summary>
        /// Procedure already defined error
        /// </summary>
        public const int ProcedureAlreadyDefinedError = 401;

        /// <summary>
        /// Procedure return at global scope error
        /// </summary>
        public const int ReturnAtGlobalScopeError = 402;

        /// <summary>
        /// Fatal error
        /// </summary>
        public const int FatalError = 500;


		/// <summary>
		/// Constructor
		/// </summary>
		internal Report()
		{
		}

		/// <summary>
		/// Report an error from the parser.
		/// </summary>
		/// <param name="location">current location</param>
		/// <param name="state">state in grammar</param>
		/// <param name="token">current token</param>
		/// <param name="tokenValue">current token value</param>
		/// <param name="expecting">list of expecting token name</param>
		public void Error(Location location, int state, int token, object tokenValue, string[] expecting)
		{
			// Lexical error
            int code = -1;
            string message = null;
            if (token == Token.ERROR)
            {
                code = Report.LexicalError;
                message = String.Format("Illegal character {0}", tokenValue.ToString());
            }

            // Unexpected end of file error
            else if (token == Token.EOF)
            {
                code = Report.UnexpectedEOFError;
                message = "Unexpected end of file";
            }

            // Syntax error
            else
            {
                StringBuilder expectings = new StringBuilder();
                int length = expecting.Length;
                for (int i = 0; i < length; i++)
                {
                    expectings.Append(' ');
                    expectings.Append(expecting[i]);
                }
                code = Report.SyntaxError;
                message = String.Format("Syntax error on \"{0}\", state {1}, expecting:{2}", tokenValue, state.ToString(), expectings.ToString());
            }
            
            // Throw error
            throw new SyntaxErrorException(message, location.SourceUnit, new SourceSpan(location.SourceLocation, location.SourceLocation), code, Severity.Error);
		}

		/// <summary>
		/// Report a fatal error.
		/// </summary>
		/// <param name="location">current location</param>
        /// <param name="errorCode">error code</param>
		/// <param name="errorText">text to qualify error</param>
		public void Error(Location location, int errorCode, string errorText)
		{
            throw new SyntaxErrorException(errorText, location.SourceUnit, new SourceSpan(location.SourceLocation, location.SourceLocation), errorCode, Severity.FatalError);
		}
	}
}

