#region MyJScript - LGPL Copyright (c) 2008 Lionel Lask
//
// This file is part of MyJScript.
//
// MyJScript is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// MyJScript is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with MyJScript; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
//================================================================================
#endregion

using System;
using System.IO;
using System.Text;
using System.Collections.Generic;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Microsoft.Scripting;
using Microsoft.Scripting.Hosting;
using MyJScript.Compiler;
using MyJScript.DLR;

namespace MyJScript.Test
{
    /// <summary>
    /// Unit test for Lexer class
    /// </summary>
    [TestClass]
    public class LexerTest
    {
        // Constructor.
        public LexerTest()
        {
        }

        /// <summary>
        /// Set up DLR domain manager
        /// </summary>
        [TestInitialize]
        public void Initialize()
        {
            Utils.Initialize();
        }

        /// <summary>
        /// Delete temp files
        /// </summary>
        [TestCleanup]
        public void Cleanup()
        {
            Utils.Cleanup();
        }

        /// <summary>
        /// Test constructor.
        /// </summary>
        [TestMethod]
        public void Constructor()
        {
            SourceUnit src = Utils.Source("fdfd");
            string v = src.GetCode();

            Lexer l1 = new Lexer(Utils.Source(null));
            Assert.IsNull(l1.value());
            Assert.IsFalse(l1.advance());
            Assert.AreEqual((int)Lexer.TokenType.EOF, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Error, l1.token());
            Assert.AreEqual(1, l1.Location.Col);
            Assert.AreEqual(1, l1.Location.Row);

            Lexer l2 = new Lexer(Utils.Source(""));
            Assert.IsNull(l2.value());
            Assert.IsFalse(l2.advance());
            Assert.AreEqual((int)Lexer.TokenType.EOF, l2.token());
            Assert.AreEqual((int)Lexer.TokenType.Error, l2.token());
            Assert.AreEqual(1, l2.Location.Col);
            Assert.AreEqual(1, l2.Location.Row);

            Lexer l3 = new Lexer(Utils.Source("    \r\n \t "));
            Assert.IsNull(l3.value());
            Assert.IsTrue(l3.advance());
            Assert.AreEqual((int)Lexer.TokenType.EOF, l3.token());
            Assert.AreEqual((int)Lexer.TokenType.Error, l3.token());
            Assert.AreEqual(4, l3.Location.Col);
            Assert.AreEqual(2, l3.Location.Row);
        }

        /// <summary>
        /// Test operators
        /// </summary>
        [TestMethod]
        public void Operators()
        {
            Lexer l1 = new Lexer(Utils.Source("+*><<====!=/->=>"));
            Assert.AreEqual((int)Lexer.TokenType.OperatorPlus, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.OperatorMult, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.OperatorGreater, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.OperatorLess, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.OperatorLessOrEqual, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.OperatorEqualEqual, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.OperatorEqual, l1.token());            
            Assert.AreEqual((int)Lexer.TokenType.OperatorNotEqual, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.OperatorDiv, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.OperatorMinus, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.OperatorGreaterOrEqual, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.OperatorGreater, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.EOF, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Error, l1.token());
            Assert.AreEqual(17, l1.Location.Col);
            Assert.AreEqual(1, l1.Location.Row);
        }

        /// <summary>
        /// Test punctuation
        /// </summary>
        [TestMethod]
        public void Punctuation()
        {
            Lexer l1 = new Lexer(Utils.Source(",;{)\r\n}("));
            Assert.AreEqual((int)Lexer.TokenType.Coma, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.SemiColon, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.LeftBra, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.RightPar, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.RightBra, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.LeftPar, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.EOF, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Error, l1.token());
            Assert.AreEqual(3, l1.Location.Col);
            Assert.AreEqual(2, l1.Location.Row);
        }

        /// <summary>
        /// Test number
        /// </summary>
        [TestMethod]
        public void Numbers()
        {           
            Lexer l1 = new Lexer(Utils.Source("12 345 007 999\r\n0 -45"));
            Assert.AreEqual((int)Lexer.TokenType.Number, l1.token());
            Assert.AreEqual(12, l1.value());
            Assert.AreEqual((int)Lexer.TokenType.Number, l1.token());
            Assert.AreEqual(345, l1.value());
            Assert.AreEqual((int)Lexer.TokenType.Number, l1.token());
            Assert.AreEqual(7, l1.value());
            Assert.AreEqual((int)Lexer.TokenType.Number, l1.token());
            Assert.AreEqual(999, l1.value());
            Assert.AreEqual((int)Lexer.TokenType.Number, l1.token());
            Assert.AreEqual(0, l1.value());
            Assert.AreEqual((int)Lexer.TokenType.OperatorMinus, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Number, l1.token());
            Assert.AreEqual(45, l1.value());
            Assert.AreEqual((int)Lexer.TokenType.EOF, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Error, l1.token());
            Assert.AreEqual(6, l1.Location.Col);
            Assert.AreEqual(2, l1.Location.Row);
        }

        /// <summary>
        /// Test string
        /// </summary>
        [TestMethod]
        public void Strings()
        {
            Lexer l1 = new Lexer(Utils.Source("'Hello World!' \"Double\tquoted\"\r\n'Double in \"single\"'\r\n\"Single in'Double'\"\r\n\"\"'A'''\r\n'NoEnd\r\n"));
            Assert.AreEqual((int)Lexer.TokenType.String, l1.token());
            Assert.AreEqual("Hello World!", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.String, l1.token());
            Assert.AreEqual("Double\tquoted", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.String, l1.token());
            Assert.AreEqual("Double in \"single\"", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.String, l1.token());
            Assert.AreEqual("Single in'Double'", l1.value());
            Assert.AreEqual(20, l1.Location.Col);
            Assert.AreEqual(3, l1.Location.Row);
            Assert.AreEqual((int)Lexer.TokenType.String, l1.token());
            Assert.AreEqual("", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.String, l1.token());
            Assert.AreEqual("A", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.String, l1.token());
            Assert.AreEqual("", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.Error, l1.token());
            Assert.AreEqual(1, l1.Location.Col);
            Assert.AreEqual(5, l1.Location.Row);
        }

        /// <summary>
        /// Test ident and keywords
        /// </summary>
        [TestMethod]
        public void IdentAndKeyword()
        {
            Lexer l1 = new Lexer(Utils.Source("while a var _x function B9+34-rtwhile\treturn if else\r\nvariable  _ *Var new using"));
            Assert.AreEqual((int)Lexer.TokenType.While, l1.token());
            Assert.AreEqual("while", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("a", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.Var, l1.token());
            Assert.AreEqual("var", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("_x", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.Function, l1.token());
            Assert.AreEqual("function", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("B9", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.OperatorPlus, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Number, l1.token());
            Assert.AreEqual(34, l1.value());
            Assert.AreEqual((int)Lexer.TokenType.OperatorMinus, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("rtwhile", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.Return, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.If, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Else, l1.token()); 
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("variable", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("_", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.OperatorMult, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("Var", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.New, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Using, l1.token());
            Assert.AreEqual(27, l1.Location.Col);
            Assert.AreEqual(2, l1.Location.Row);
        }

        /// <summary>
        /// Test Error
        /// </summary>
        [TestMethod]
        public void Error()
        {
            Lexer l1 = new Lexer(Utils.Source("+12~i"));
            Assert.AreEqual((int)Lexer.TokenType.OperatorPlus, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Number, l1.token());
            Assert.AreEqual(12, l1.value());
            Assert.AreEqual((int)Lexer.TokenType.Error, l1.token());
            Assert.AreEqual("~", l1.value().ToString());
        }

        /// <summary>
        /// Test full MJS program
        /// </summary>
        [TestMethod]
        public void Program()
        {
            Lexer l1 = new Lexer(Utils.Source(@"
                function twotime(n)
                {
                    return 2*n;
                }

                var i = 12;

                write(twotime(i));
            "));
            Assert.AreEqual((int)Lexer.TokenType.Function, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("twotime", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.LeftPar, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("n", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.RightPar, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.LeftBra, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Return, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Number, l1.token());
            Assert.AreEqual(2, l1.value());
            Assert.AreEqual((int)Lexer.TokenType.OperatorMult, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("n", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.SemiColon, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.RightBra, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Var, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("i", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.OperatorEqual, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Number, l1.token());
            Assert.AreEqual(12, l1.value());
            Assert.AreEqual((int)Lexer.TokenType.SemiColon, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("write", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.LeftPar, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("twotime", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.LeftPar, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.Ident, l1.token());
            Assert.AreEqual("i", l1.value());
            Assert.AreEqual((int)Lexer.TokenType.RightPar, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.RightPar, l1.token());
            Assert.AreEqual((int)Lexer.TokenType.SemiColon, l1.token());
            Assert.AreEqual(35, l1.Location.Col);
            Assert.AreEqual(9, l1.Location.Row);
            Assert.AreEqual((int)Lexer.TokenType.EOF, l1.token());
            Assert.IsFalse(l1.advance());
            Assert.AreEqual((int)Lexer.TokenType.Error, l1.token());
        }
    }
}
